/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * File: devices.h
 * Created by: luke <luke@dhcp-45-369>
 * Created on: Mon Feb 24 21:51:18 2003
 */

#include "devices.h"
#include "exec.h"
#include "gbcommon.h"
#include <glib/gprintf.h>

DeviceData *devices_new_devicedata();
void devices_delete_devicedata(DeviceData* dev);
void devices_add_device(DeviceData* dev);
void devices_clear_devicedata();
gboolean devices_parse_cdrecord_output(gchar* buffer, const gchar* busname);


GList *devices = NULL;

const gint nobusses = 3;
const gchar* busses[3] = 
{
	"SCSI",
	"ATAPI",
	"ATA"
};


gboolean
devices_init()
{
	GB_LOG_FUNC
	devices = NULL;
	return devices_probescsibus();
}


DeviceData *
devices_new_devicedata()
{
	GB_LOG_FUNC
	DeviceData* dev = (DeviceData*)g_new(DeviceData, 1);

	memset(dev->scsidevice, 0, 8);
	memset(dev->displayname, 0, 256);

	return dev;
}


void
devices_delete_devicedata(DeviceData * dev)
{
	GB_LOG_FUNC
	g_free(dev);
}


void
devices_add_device(DeviceData * dev)
{
	GB_LOG_FUNC
	if(!devices)
		devices = g_list_append(devices, dev);
	else
		g_list_append(devices, dev);

	g_message("devices_add_device - Added [%s] [%s]", dev->displayname, dev->scsidevice);
}


void
devices_populate_optionmenu(GtkWidget* option_menu)
{
	GB_LOG_FUNC
	g_return_if_fail(option_menu != NULL);
																																		   
	GtkWidget* menu = gtk_option_menu_get_menu(GTK_OPTION_MENU(option_menu));
	if(menu != NULL)
		gtk_widget_destroy(menu);
	menu = gtk_menu_new();
	gtk_widget_show(menu);
																																		   
	GList *items = devices;	
	while(items)
	{
		GtkWidget* item = gtk_menu_item_new_with_label(
			((DeviceData *)items->data)->displayname);
		gtk_widget_show(item);
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
		items = items->next;
	}
	
	gtk_option_menu_set_history(GTK_OPTION_MENU(option_menu), 0);																																		  
	gtk_option_menu_set_menu(GTK_OPTION_MENU(option_menu), menu);	
}


gint
devices_get_index_from_scsiid(const gchar* scsiid)
{
	GB_LOG_FUNC
	g_return_val_if_fail(scsiid != NULL, 0);
	
	gint index = 0;
	GList *items = devices;

	while(items)
	{
		DeviceData* dev = (DeviceData*)items->data;
		if(g_ascii_strcasecmp(dev->scsidevice, scsiid) == 0)
			break;

		items = items->next;
		index++;
	}
	
	return index;
}


const gchar* 
devices_get_scsiid_from_index(const gint index)
{
	GB_LOG_FUNC
	gchar* scsiid = NULL;
	DeviceData* data = g_list_nth_data(devices, index);
	if(data != NULL)
		scsiid = data->scsidevice;
	
	return scsiid;
}


void
devices_clear_devicedata()
{
	GB_LOG_FUNC

	GList *items = devices;

	while(items)
	{
		g_message("devices_clear_devicedata - removing [%s]",
		  ((DeviceData *) items->data)->scsidevice);
		devices_delete_devicedata((DeviceData *) items->data);
		items = items->next;
	}

	g_list_free(devices);
	devices = NULL;
}


gboolean
devices_parse_cdrecord_output(gchar* buffer, const gchar* busname)
{
	GB_LOG_FUNC
	g_return_val_if_fail(buffer != NULL, FALSE);
	gboolean ok = TRUE;		
	
	gint bus = 0;	
	for(; bus != 16; bus++)
	{				
		gchar* scsibusses = g_strdup_printf("scsibus%d", bus);
	
		gchar *ptr = strstr(buffer, scsibusses);
		if(ptr != NULL)
		{
			static const gint devices = 8;
			gint i = 0;
			for(; i < devices; i++)
			{
				gchar* device = g_strdup_printf("%d)", i);
	
				if((ptr = strstr(ptr, device)))
				{
					ptr += strlen(device) + 1;
					if((*ptr) != '*')
					{
						DeviceData *dev = devices_new_devicedata();
	
						/* Parse the display name and remove the 's */
						gchar *tmpptr = ptr;
						while(*tmpptr != '\n')
						{
							if(*tmpptr != '\'')
								strncat(dev->displayname, tmpptr, 1);
							
							tmpptr++;
						}
						
						/* add on the bus so we know what we are using */
						strcat(dev->displayname, " (");
						strcat(dev->displayname, busname);
						strcat(dev->displayname, ")");						
	
						/* Copy the bus id stuff ie 0,0,0 to the device struct. 
						   We append ATAPI only as the cdrecord command line
						   assumes SCSI. */
						if(g_ascii_strcasecmp(busname, busses[0]) != 0)
							g_sprintf(dev->scsidevice, "%s:", busname);
						
						strncat(dev->scsidevice, ptr - 11, 5);
						/*dev->scsidevice[5] = '\0';*/
	
						/* Add it to the device collection */
						devices_add_device(dev);
					}
				}
				g_free(device);
			}
		}
		g_free(scsibusses);
	}

	return ok;
}


gboolean
devices_probescsibus()
{
	GB_LOG_FUNC

	gboolean ok = TRUE;

	devices_clear_devicedata();	
	
	gint i = 0;
	for(; i < nobusses && ok; i++)
	{
		gchar command[32] = "cdrecord -scanbus";
		if(g_ascii_strcasecmp(busses[i], busses[0]) != 0)
		{
			strcat(command, " dev=");
			strcat(command, busses[i]);
		}
		
		GString* buffer = exec_run_cmd(command);
		if(buffer == NULL)
		{
			g_critical("devices_probescsibus - Failed to scan the scsi bus");
			ok = FALSE;
		}
		else if(!devices_parse_cdrecord_output(buffer->str, busses[i]))
		{
			g_critical("devices_probescsibus - failed to parse cdrecord output");
			ok = FALSE;
		}
		
		if(buffer != NULL)
			g_string_free(buffer, TRUE);
	}

	return ok;
}


const gchar* 
devices_guess_cdrom()
{
	GB_LOG_FUNC
	const gchar* cdrom = NULL;	
	GList *items = devices;

	while(items)
	{
		DeviceData *dev =(DeviceData *) items->data;
		if(strstr(dev->displayname, "CD-ROM") != NULL)
		{
			cdrom = dev->displayname;
			break;
		}
		items = items->next;		
	}
	
	return cdrom;
}
